# torchcell/pypy_adapters/kuzmin2018_pypy_adapter.py
# [[torchcellpy_adapters.kuzmin2018_pypy_adapter]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/pypy_adapters/kuzmin2018_pypy_adapter.py
# Test file: tests/torchcell/pypy_adapters/test_kuzmin2018_pypy_adapter.py

from tqdm import tqdm
import hashlib
import json
from biocypher._create import BioCypherEdge, BioCypherNode
from biocypher._logger import logger
from typing import Generator, Set
from torchcell.datamodels import BaseGenotype, InterferenceGenotype, DeletionGenotype
from concurrent.futures import ProcessPoolExecutor, as_completed
from biocypher import BioCypher

from torchcell.dataset_readers import LmdbDatasetReader

logger.debug(f"Loading module {__name__}.")


class SmfKuzmin2018Adapter:
    def __init__(self, dataset: LmdbDatasetReader):
        self.dataset = dataset

    def get_nodes(self) -> None:
        logger.info("Getting nodes.")
        logger.info("Get experiment reference nodes.")
        yield from self._get_experiment_reference_nodes()
        logger.info("Get genome nodes.")
        yield from self._get_genome_nodes()
        logger.info("Get experiment nodes.")
        yield from self._get_experiment_nodes()
        logger.info("Get dataset nodes.")
        yield from self._get_dataset_nodes()
        logger.info("Get genotype nodes.")
        logger.info("--- perturbation nodes.")
        yield from self._get_genotype_nodes()
        logger.info("Get environment nodes.")
        yield from self._get_environment_nodes()
        logger.info("Get media nodes.")
        yield from self._get_media_nodes()
        logger.info("Get temperature nodes.")
        yield from self._get_temperature_nodes()
        logger.info("Get phenotype nodes.")
        yield from self._get_phenotype_nodes()

    def _get_experiment_reference_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherNode(
                node_id=experiment_ref_id,
                preferred_id=f"SmfKuzmin2018_reference_{i}",
                node_label="experiment reference",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data.reference.model_dump()),
                },
            )

    def _get_genome_nodes(self) -> None:
        seen_node_ids: Set[str] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
            ).hexdigest()

            if genome_id not in seen_node_ids:
                seen_node_ids.add(genome_id)
                yield BioCypherNode(
                    node_id=genome_id,
                    preferred_id=f"reference_genome_{i}",
                    node_label="genome",
                    properties={
                        "species": data.reference.reference_genome.species,
                        "strain": data.reference.reference_genome.strain,
                        "serialized_data": json.dumps(
                            data.reference.reference_genome.model_dump()
                        ),
                    },
                )

    def _get_experiment_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()

            yield BioCypherNode(
                node_id=experiment_id,
                preferred_id=f"SmfKuzmin2018_{i}",
                node_label="experiment",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data["experiment"].model_dump()),
                },
            )

    def _get_genotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            genotype_id = hashlib.md5(
                json.dumps(data["experiment"].genotype.model_dump()).encode("utf-8")
            ).hexdigest()

            if genotype_id not in seen_node_ids:
                seen_node_ids.add(genotype_id)
                systematic_gene_name = data[
                    "experiment"
                ].genotype.perturbation.systematic_gene_name
                perturbed_gene_name = data[
                    "experiment"
                ].genotype.perturbation.perturbed_gene_name
                description = data["experiment"].genotype.perturbation.description
                perturbation_type = data[
                    "experiment"
                ].genotype.perturbation.perturbation_type

                self._get_perturbation(data["experiment"].genotype)

                yield BioCypherNode(
                    node_id=genotype_id,
                    preferred_id=f"genotype_{i}",
                    node_label="genotype",
                    properties={
                        "systematic_gene_names": [systematic_gene_name],
                        "perturbed_gene_names": [perturbed_gene_name],
                        "is_deletion_genotype": isinstance(
                            data["experiment"].genotype, DeletionGenotype
                        ),
                        "is_interference_genotype": isinstance(
                            data["experiment"].genotype, InterferenceGenotype
                        ),
                        "description": description,
                        "perturbation_types": [perturbation_type],
                        "serialized_data": json.dumps(
                            data["experiment"].genotype.model_dump()
                        ),
                    },
                )

    @staticmethod
    def _get_perturbation(
        genotype: BaseGenotype,
    ) -> Generator[BioCypherNode, None, None]:
        perturbation_id = hashlib.md5(
            json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
        ).hexdigest()

        yield BioCypherNode(
            node_id=perturbation_id,
            preferred_id=genotype.perturbation.perturbation_type,
            node_label="perturbation",
            properties={
                "systematic_gene_name": [
                    genotype.perturbation.systematic_gene_name
                ],
                "perturbed_gene_name": [genotype.perturbation.perturbed_gene_name],
                "description": genotype.perturbation.description,
                "perturbation_type": genotype.perturbation.perturbation_type,
                "strain_id": genotype.perturbation.strain_id,
                "serialized_data": json.dumps(genotype.perturbation.model_dump()),
            },
        )

    def _get_environment_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(data["experiment"].environment.media.model_dump())

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data["experiment"].environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["reference"].reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                )

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.model_dump()
                        ),
                    },
                )

    def _get_media_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["experiment"].environment.media.name
                state = data["experiment"].environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.media.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["reference"].reference_environment.media.name
                state = data["reference"].reference_environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.media.model_dump()
                        ),
                    },
                )

    def _get_temperature_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data["experiment"].environment.temperature.value,
                        "unit": data["experiment"].environment.temperature.unit,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.temperature.model_dump()
                        ),
                    },
                )

        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "description": data[
                            "reference"
                        ].reference_environment.temperature.description,
                        "serialized_data": json.dumps(
                            data[
                                "reference"
                            ].reference_environment.temperature.model_dump()
                        ),
                    },
                )

    def _get_phenotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["experiment"].phenotype.graph_level
                label = data["experiment"].phenotype.label
                label_error = data["experiment"].phenotype.label_error
                fitness = data["experiment"].phenotype.fitness
                fitness_std = data["experiment"].phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["experiment"].phenotype.model_dump()
                        ),
                    },
                )

        # References
        for i, data in tqdm(enumerate(self.dataset)):
            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data["reference"].reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["reference"].reference_phenotype.graph_level
                label = data["reference"].reference_phenotype.label
                label_error = data["reference"].reference_phenotype.label_error
                fitness = data["reference"].reference_phenotype.fitness
                fitness_std = data["reference"].reference_phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["reference"].reference_phenotype.model_dump()
                        ),
                    },
                )

    def _get_dataset_nodes(self) -> None:
        yield BioCypherNode(
            node_id="SmfKuzmin2018",
            preferred_id="SmfKuzmin2018",
            node_label="dataset",
        )

    def get_edges(self) -> None:
        logger.info("Generating edges.")
        logger.info("Get dataset experiment reference edges.")
        yield from self._get_dataset_experiment_ref_edges()
        logger.info("Get experiment dataset edges.")
        yield from self._get_experiment_dataset_edges()
        logger.info("Get experiment reference experiment edges.")
        yield from self._get_experiment_ref_experiment_edges()
        logger.info("Get genotype experiment edges.")
        logger.info("--- perturbation genotype edges.")
        yield from self._get_genotype_experiment_edges()
        logger.info("Get environment experiment edges.")
        yield from self._get_environment_experiment_edges()
        logger.info("Get environment experiment reference edges.")
        yield from self._get_environment_experiment_ref_edges()
        logger.info("Get phenotype experiment edges.")
        yield from self._get_phenotype_experiment_edges()
        logger.info("Get phenotype experiment reference edges.")
        yield from self._get_phenotype_experiment_ref_edges()
        logger.info("Get media environment edges.")
        yield from self._get_media_environment_edges()
        logger.info("Get temperature environment edges.")
        yield from self._get_temperature_environment_edges()
        logger.info("Get genome experiment reference edges.")
        yield from self._get_genome_edges()

    def _get_dataset_experiment_ref_edges(self):
        # concept level
        for data in tqdm(self.dataset.experiment_reference_index):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_ref_id,
                target_id="SmfKuzmin2018",
                relationship_label="experiment reference member of",
            )

    def _get_experiment_dataset_edges(self):
        # concept level
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_id,
                target_id="SmfKuzmin2018",
                relationship_label="experiment member of",
            )

    def _get_experiment_ref_experiment_edges(self):
        # instance level
        for data in tqdm(self.dataset.experiment_reference_index):
            dataset_subset = self.dataset[data.index]
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            for i, data in enumerate(dataset_subset):
                experiment_id = hashlib.md5(
                    json.dumps(data["experiment"].model_dump()).encode("utf-8")
                ).hexdigest()
                yield BioCypherEdge(
                    source_id=experiment_ref_id,
                    target_id=experiment_id,
                    relationship_label="experiment reference of",
                )

    def _get_genotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            genotype_id = hashlib.md5(
                json.dumps(data["experiment"].genotype.model_dump()).encode("utf-8")
            ).hexdigest()

            self._get_perturbation_genotype_edges(
                genotype=data["experiment"].genotype, genotype_id=genotype_id
            )
            yield BioCypherEdge(
                source_id=genotype_id,
                target_id=experiment_id,
                relationship_label="genotype member of",
            )

    @staticmethod
    def _get_perturbation_genotype_edges(
        genotype: BaseGenotype, genotype_id: str
    ) -> Generator[BioCypherEdge, None, None]:
        perturbation_id = hashlib.md5(
            json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
        ).hexdigest()

        yield BioCypherEdge(
            source_id=perturbation_id,
            target_id=genotype_id,
            relationship_label="perturbation member of",
        )

    def _get_environment_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_pairs: Set[tuple] = set()
        # Linking environments to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            env_experiment_pair = (environment_id, experiment_id)
            if env_experiment_pair not in seen_environment_experiment_pairs:
                seen_environment_experiment_pairs.add(env_experiment_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_id,
                    relationship_label="environment member of",
                )

    def _get_environment_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_ref_pairs: Set[tuple] = set()

        # Linking environments to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            environment_id = hashlib.md5(
                json.dumps(data.reference.reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            env_experiment_ref_pair = (environment_id, experiment_ref_id)
            if env_experiment_ref_pair not in seen_environment_experiment_ref_pairs:
                seen_environment_experiment_ref_pairs.add(env_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_ref_id,
                    relationship_label="environment member of",
                )

    def _get_phenotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_pairs: Set[tuple] = set()

        # Linking phenotypes to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            phenotype_experiment_pair = (phenotype_id, experiment_id)
            if phenotype_experiment_pair not in seen_phenotype_experiment_pairs:
                seen_phenotype_experiment_pairs.add(phenotype_experiment_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_id,
                    relationship_label="phenotype member of",
                )

    def _get_phenotype_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_ref_pairs: Set[tuple] = set()

        # Linking phenotypes to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data.reference.reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            phenotype_experiment_ref_pair = (phenotype_id, experiment_ref_id)
            if phenotype_experiment_ref_pair not in seen_phenotype_experiment_ref_pairs:
                seen_phenotype_experiment_ref_pairs.add(phenotype_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_ref_id,
                    relationship_label="phenotype member of",
                )

    def _get_media_environment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_media_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            media_environment_pair = (media_id, environment_id)
            if media_environment_pair not in seen_media_environment_pairs:
                seen_media_environment_pairs.add(media_environment_pair)

                yield BioCypherEdge(
                    source_id=media_id,
                    target_id=environment_id,
                    relationship_label="media member of",
                )

    def _get_temperature_environment_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_temperature_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            temperature_environment_pair = (temperature_id, environment_id)
            if temperature_environment_pair not in seen_temperature_environment_pairs:
                seen_temperature_environment_pairs.add(temperature_environment_pair)

                yield BioCypherEdge(
                    source_id=temperature_id,
                    target_id=environment_id,
                    relationship_label="temperature member of",
                )

    def _get_genome_edges(self) -> None:
        seen_genome_experiment_ref_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            genome_experiment_ref_pair = (genome_id, experiment_ref_id)
            if genome_experiment_ref_pair not in seen_genome_experiment_ref_pairs:
                seen_genome_experiment_ref_pairs.add(genome_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=genome_id,
                    target_id=experiment_ref_id,
                    relationship_label="genome member of",
                )


class DmfKuzmin2018Adapter:
    def __init__(self, dataset: LmdbDatasetReader):
        self.dataset = dataset

    def get_nodes(self) -> None:
        logger.info("Getting nodes.")
        logger.info("Get experiment reference nodes.")
        yield from self._get_experiment_reference_nodes()
        logger.info("Get genome nodes.")
        yield from self._get_genome_nodes()
        logger.info("Get experiment nodes.")
        yield from self._get_experiment_nodes()
        logger.info("Get dataset nodes.")
        yield from self._get_dataset_nodes()
        logger.info("Get genotype nodes.")
        logger.info("--- perturbation nodes.")
        yield from self._get_genotype_nodes()
        logger.info("Get environment nodes.")
        yield from self._get_environment_nodes()
        logger.info("Get media nodes.")
        yield from self._get_media_nodes()
        logger.info("Get temperature nodes.")
        yield from self._get_temperature_nodes()
        logger.info("Get phenotype nodes.")
        yield from self._get_phenotype_nodes()

    def _get_experiment_reference_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherNode(
                node_id=experiment_ref_id,
                preferred_id=f"DmfKuzmin2018_reference_{i}",
                node_label="experiment reference",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data.reference.model_dump()),
                },
            )

    def _get_genome_nodes(self) -> None:
        seen_node_ids: Set[str] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
            ).hexdigest()

            if genome_id not in seen_node_ids:
                seen_node_ids.add(genome_id)
                yield BioCypherNode(
                    node_id=genome_id,
                    preferred_id=f"reference_genome_{i}",
                    node_label="genome",
                    properties={
                        "species": data.reference.reference_genome.species,
                        "strain": data.reference.reference_genome.strain,
                        "serialized_data": json.dumps(
                            data.reference.reference_genome.model_dump()
                        ),
                    },
                )

    def _get_experiment_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()

            yield BioCypherNode(
                node_id=experiment_id,
                preferred_id=f"DmfKuzmin2018_{i}",
                node_label="experiment",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data["experiment"].model_dump()),
                },
            )

    def _get_genotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            for genotype in data["experiment"].genotype:
                genotype_id = hashlib.md5(
                    json.dumps(genotype.model_dump()).encode("utf-8")
                ).hexdigest()

                if genotype_id not in seen_node_ids:
                    seen_node_ids.add(genotype_id)
                    systematic_gene_name = genotype.perturbation.systematic_gene_name
                    perturbed_gene_name = genotype.perturbation.perturbed_gene_name
                    description = genotype.perturbation.description
                    perturbation_type = genotype.perturbation.perturbation_type
                    self._get_perturbation(genotype)

                    yield BioCypherNode(
                        node_id=genotype_id,
                        preferred_id=f"genotype_{i}",
                        node_label="genotype",
                        properties={
                            "systematic_gene_names": [systematic_gene_name],
                            "perturbed_gene_names": [perturbed_gene_name],
                            "is_deletion_genotype": isinstance(
                                data["experiment"].genotype, DeletionGenotype
                            ),
                            "is_interference_genotype": isinstance(
                                data["experiment"].genotype, InterferenceGenotype
                            ),
                            "description": description,
                            "perturbation_types": [perturbation_type],
                            "serialized_data": json.dumps(genotype.model_dump()),
                        },
                    )

    @staticmethod
    def _get_perturbation(
        genotype: BaseGenotype,
    ) -> Generator[BioCypherNode, None, None]:
        perturbation_id = hashlib.md5(
            json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
        ).hexdigest()

        yield BioCypherNode(
            node_id=perturbation_id,
            preferred_id=genotype.perturbation.perturbation_type,
            node_label="perturbation",
            properties={
                "systematic_gene_name": [
                    genotype.perturbation.systematic_gene_name
                ],
                "perturbed_gene_name": [genotype.perturbation.perturbed_gene_name],
                "description": genotype.perturbation.description,
                "perturbation_type": genotype.perturbation.perturbation_type,
                "strain_id": genotype.perturbation.strain_id,
                "serialized_data": json.dumps(genotype.perturbation.model_dump()),
            },
        )

    def _get_environment_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(data["experiment"].environment.media.model_dump())

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data["experiment"].environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["reference"].reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                )

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.model_dump()
                        ),
                    },
                )

    def _get_media_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["experiment"].environment.media.name
                state = data["experiment"].environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.media.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["reference"].reference_environment.media.name
                state = data["reference"].reference_environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.media.model_dump()
                        ),
                    },
                )

    def _get_temperature_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data["experiment"].environment.temperature.value,
                        "unit": data["experiment"].environment.temperature.unit,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.temperature.model_dump()
                        ),
                    },
                )

        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "description": data[
                            "reference"
                        ].reference_environment.temperature.description,
                        "serialized_data": json.dumps(
                            data[
                                "reference"
                            ].reference_environment.temperature.model_dump()
                        ),
                    },
                )

    def _get_phenotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["experiment"].phenotype.graph_level
                label = data["experiment"].phenotype.label
                label_error = data["experiment"].phenotype.label_error
                fitness = data["experiment"].phenotype.fitness
                fitness_std = data["experiment"].phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["experiment"].phenotype.model_dump()
                        ),
                    },
                )

        # References
        for i, data in tqdm(enumerate(self.dataset)):
            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data["reference"].reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["reference"].reference_phenotype.graph_level
                label = data["reference"].reference_phenotype.label
                label_error = data["reference"].reference_phenotype.label_error
                fitness = data["reference"].reference_phenotype.fitness
                fitness_std = data["reference"].reference_phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["reference"].reference_phenotype.model_dump()
                        ),
                    },
                )

    def _get_dataset_nodes(self) -> None:
        yield BioCypherNode(
            node_id="DmfKuzmin2018",
            preferred_id="DmfKuzmin2018",
            node_label="dataset",
        )

    def get_edges(self) -> None:
        logger.info("Generating edges.")
        logger.info("Get dataset experiment reference edges.")
        yield from self._get_dataset_experiment_ref_edges()
        logger.info("Get experiment dataset edges.")
        yield from self._get_experiment_dataset_edges()
        logger.info("Get experiment reference experiment edges.")
        yield from self._get_experiment_ref_experiment_edges()
        logger.info("Get genotype experiment edges.")
        logger.info("--- perturbation genotype edges.")
        yield from self._get_genotype_experiment_edges()
        logger.info("Get environment experiment edges.")
        yield from self._get_environment_experiment_edges()
        logger.info("Get environment experiment reference edges.")
        yield from self._get_environment_experiment_ref_edges()
        logger.info("Get phenotype experiment edges.")
        yield from self._get_phenotype_experiment_edges()
        logger.info("Get phenotype experiment reference edges.")
        yield from self._get_phenotype_experiment_ref_edges()
        logger.info("Get media environment edges.")
        yield from self._get_media_environment_edges()
        logger.info("Get temperature environment edges.")
        yield from self._get_temperature_environment_edges()
        logger.info("Get genome experiment reference edges.")
        yield from self._get_genome_edges()

    def _get_dataset_experiment_ref_edges(self):
        # concept level
        for data in tqdm(self.dataset.experiment_reference_index):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_ref_id,
                target_id="DmfKuzmin2018",
                relationship_label="experiment reference member of",
            )

    def _get_experiment_dataset_edges(self):
        # concept level
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_id,
                target_id="DmfKuzmin2018",
                relationship_label="experiment member of",
            )

    def _get_experiment_ref_experiment_edges(self):
        # instance level
        print()
        for data in tqdm(self.dataset.experiment_reference_index):
            dataset_subset = self.dataset[data.index]
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            for i, data in enumerate(dataset_subset):
                experiment_id = hashlib.md5(
                    json.dumps(data["experiment"].model_dump()).encode("utf-8")
                ).hexdigest()
                yield BioCypherEdge(
                    source_id=experiment_ref_id,
                    target_id=experiment_id,
                    relationship_label="experiment reference of",
                )

    def _get_genotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        # CHECK if needed - don't think needed since exp ref index
        # seen_genotype_experiment_pairs: Set[tuple] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            for genotype in data["experiment"].genotype:
                genotype_id = hashlib.md5(
                    json.dumps(genotype.model_dump()).encode("utf-8")
                ).hexdigest()

                self._get_perturbation_genotype_edges(
                    genotype=genotype, genotype_id=genotype_id
                )

                # CHECK if needed - don't think needed since exp ref index
                # genotype_experiment_pair = (genotype_id, experiment_id)
                # if genotype_experiment_pair not in seen_genotype_experiment_pairs:
                #     seen_genotype_experiment_pairs.add(genotype_experiment_pair)

                yield BioCypherEdge(
                    source_id=genotype_id,
                    target_id=experiment_id,
                    relationship_label="genotype member of",
                )

    @staticmethod
    def _get_perturbation_genotype_edges(
        genotype: BaseGenotype, genotype_id: str
    ) -> Generator[BioCypherEdge, None, None]:
        if genotype.perturbation:
            perturbation_id = hashlib.md5(
                json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
            ).hexdigest()

            yield BioCypherEdge(
                source_id=perturbation_id,
                target_id=genotype_id,
                relationship_label="perturbation member of",
            )

    def _get_environment_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_pairs: Set[tuple] = set()

        # Linking environments to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            env_experiment_pair = (environment_id, experiment_id)
            if env_experiment_pair not in seen_environment_experiment_pairs:
                seen_environment_experiment_pairs.add(env_experiment_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_id,
                    relationship_label="environment member of",
                )

    def _get_environment_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_ref_pairs: Set[tuple] = set()

        # Linking environments to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            environment_id = hashlib.md5(
                json.dumps(data.reference.reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            env_experiment_ref_pair = (environment_id, experiment_ref_id)
            if env_experiment_ref_pair not in seen_environment_experiment_ref_pairs:
                seen_environment_experiment_ref_pairs.add(env_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_ref_id,
                    relationship_label="environment member of",
                )

    def _get_phenotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_pairs: Set[tuple] = set()

        # Linking phenotypes to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            phenotype_experiment_pair = (phenotype_id, experiment_id)
            if phenotype_experiment_pair not in seen_phenotype_experiment_pairs:
                seen_phenotype_experiment_pairs.add(phenotype_experiment_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_id,
                    relationship_label="phenotype member of",
                )

    def _get_phenotype_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_ref_pairs: Set[tuple] = set()

        # Linking phenotypes to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data.reference.reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            phenotype_experiment_ref_pair = (phenotype_id, experiment_ref_id)
            if phenotype_experiment_ref_pair not in seen_phenotype_experiment_ref_pairs:
                seen_phenotype_experiment_ref_pairs.add(phenotype_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_ref_id,
                    relationship_label="phenotype member of",
                )

    def _get_media_environment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_media_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            media_environment_pair = (media_id, environment_id)
            if media_environment_pair not in seen_media_environment_pairs:
                seen_media_environment_pairs.add(media_environment_pair)

                yield BioCypherEdge(
                    source_id=media_id,
                    target_id=environment_id,
                    relationship_label="media member of",
                )

    def _get_temperature_environment_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_temperature_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            temperature_environment_pair = (temperature_id, environment_id)
            if temperature_environment_pair not in seen_temperature_environment_pairs:
                seen_temperature_environment_pairs.add(temperature_environment_pair)

                yield BioCypherEdge(
                    source_id=temperature_id,
                    target_id=environment_id,
                    relationship_label="temperature member of",
                )

    def _get_genome_edges(self) -> None:
        seen_genome_experiment_ref_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            genome_experiment_ref_pair = (genome_id, experiment_ref_id)
            if genome_experiment_ref_pair not in seen_genome_experiment_ref_pairs:
                seen_genome_experiment_ref_pairs.add(genome_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=genome_id,
                    target_id=experiment_ref_id,
                    relationship_label="genome member of",
                )


class TmfKuzmin2018Adapter:
    def __init__(self, dataset: LmdbDatasetReader):
        self.dataset = dataset

    def get_nodes(self) -> None:
        logger.info("Getting nodes.")
        logger.info("Get experiment reference nodes.")
        yield from self._get_experiment_reference_nodes()
        logger.info("Get genome nodes.")
        yield from self._get_genome_nodes()
        logger.info("Get experiment nodes.")
        yield from self._get_experiment_nodes()
        logger.info("Get dataset nodes.")
        yield from self._get_dataset_nodes()
        logger.info("Get genotype nodes.")
        logger.info("--- perturbation nodes.")
        yield from self._get_genotype_nodes()
        logger.info("Get environment nodes.")
        yield from self._get_environment_nodes()
        logger.info("Get media nodes.")
        yield from self._get_media_nodes()
        logger.info("Get temperature nodes.")
        yield from self._get_temperature_nodes()
        logger.info("Get phenotype nodes.")
        yield from self._get_phenotype_nodes()

    def _get_experiment_reference_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherNode(
                node_id=experiment_ref_id,
                preferred_id=f"TmfKuzmin2018_reference_{i}",
                node_label="experiment reference",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data.reference.model_dump()),
                },
            )

    def _get_genome_nodes(self) -> None:
        seen_node_ids: Set[str] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
            ).hexdigest()

            if genome_id not in seen_node_ids:
                seen_node_ids.add(genome_id)
                yield BioCypherNode(
                    node_id=genome_id,
                    preferred_id=f"reference_genome_{i}",
                    node_label="genome",
                    properties={
                        "species": data.reference.reference_genome.species,
                        "strain": data.reference.reference_genome.strain,
                        "serialized_data": json.dumps(
                            data.reference.reference_genome.model_dump()
                        ),
                    },
                )

    def _get_experiment_nodes(self) -> None:
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()

            yield BioCypherNode(
                node_id=experiment_id,
                preferred_id=f"TmfKuzmin2018_{i}",
                node_label="experiment",
                properties={
                    "dataset_index": i,
                    "serialized_data": json.dumps(data["experiment"].model_dump()),
                },
            )

    def _get_genotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            for genotype in data["experiment"].genotype:
                genotype_id = hashlib.md5(
                    json.dumps(genotype.model_dump()).encode("utf-8")
                ).hexdigest()

                if genotype_id not in seen_node_ids:
                    seen_node_ids.add(genotype_id)
                    systematic_gene_name = genotype.perturbation.systematic_gene_name
                    perturbed_gene_name = genotype.perturbation.perturbed_gene_name
                    description = genotype.perturbation.description
                    perturbation_type = genotype.perturbation.perturbation_type
                    self._get_perturbation(genotype)

                    yield BioCypherNode(
                        node_id=genotype_id,
                        preferred_id=f"genotype_{i}",
                        node_label="genotype",
                        properties={
                            "systematic_gene_names": [systematic_gene_name],
                            "perturbed_gene_names": [perturbed_gene_name],
                            "is_deletion_genotype": isinstance(
                                data["experiment"].genotype, DeletionGenotype
                            ),
                            "is_interference_genotype": isinstance(
                                data["experiment"].genotype, InterferenceGenotype
                            ),
                            "description": description,
                            "perturbation_types": [perturbation_type],
                            "serialized_data": json.dumps(genotype.model_dump()),
                        },
                    )

    @staticmethod
    def _get_perturbation(
        genotype: BaseGenotype,
    ) -> Generator[BioCypherNode, None, None]:
        perturbation_id = hashlib.md5(
            json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
        ).hexdigest()

        yield BioCypherNode(
            node_id=perturbation_id,
            preferred_id=genotype.perturbation.perturbation_type,
            node_label="perturbation",
            properties={
                "systematic_gene_name": [
                    genotype.perturbation.systematic_gene_name
                ],
                "perturbed_gene_name": [genotype.perturbation.perturbed_gene_name],
                "description": genotype.perturbation.description,
                "perturbation_type": genotype.perturbation.perturbation_type,
                "strain_id": genotype.perturbation.strain_id,
                "serialized_data": json.dumps(genotype.perturbation.model_dump()),
            },
        )

    def _get_environment_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(data["experiment"].environment.media.model_dump())

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data["experiment"].environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["reference"].reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            node_id = environment_id

            if node_id not in seen_node_ids:
                seen_node_ids.add(node_id)
                media = json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                )

                yield BioCypherNode(
                    node_id=node_id,
                    preferred_id=f"environment_{i}",
                    node_label="environment",
                    properties={
                        "temperature": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.model_dump()
                        ),
                    },
                )

    def _get_media_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["experiment"].environment.media.name
                state = data["experiment"].environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.media.model_dump()
                        ),
                    },
                )
        for i, data in tqdm(enumerate(self.dataset)):
            media_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.media.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data["reference"].reference_environment.media.name
                state = data["reference"].reference_environment.media.state

                yield BioCypherNode(
                    node_id=media_id,
                    preferred_id=f"media_{media_id}",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data["reference"].reference_environment.media.model_dump()
                        ),
                    },
                )

    def _get_temperature_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data["experiment"].environment.temperature.value,
                        "unit": data["experiment"].environment.temperature.unit,
                        "serialized_data": json.dumps(
                            data["experiment"].environment.temperature.model_dump()
                        ),
                    },
                )

        for i, data in tqdm(enumerate(self.dataset)):
            temperature_id = hashlib.md5(
                json.dumps(
                    data["reference"].reference_environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)

                yield BioCypherNode(
                    node_id=temperature_id,
                    preferred_id=f"temperature_{temperature_id}",
                    node_label="temperature",
                    properties={
                        "value": data[
                            "reference"
                        ].reference_environment.temperature.value,
                        "description": data[
                            "reference"
                        ].reference_environment.temperature.description,
                        "serialized_data": json.dumps(
                            data[
                                "reference"
                            ].reference_environment.temperature.model_dump()
                        ),
                    },
                )

    def _get_phenotype_nodes(self) -> Generator[BioCypherNode, None, None]:
        seen_node_ids: Set[str] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["experiment"].phenotype.graph_level
                label = data["experiment"].phenotype.label
                label_error = data["experiment"].phenotype.label_error
                fitness = data["experiment"].phenotype.fitness
                fitness_std = data["experiment"].phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["experiment"].phenotype.model_dump()
                        ),
                    },
                )

        # References
        for i, data in tqdm(enumerate(self.dataset)):
            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data["reference"].reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            if phenotype_id not in seen_node_ids:
                seen_node_ids.add(phenotype_id)
                graph_level = data["reference"].reference_phenotype.graph_level
                label = data["reference"].reference_phenotype.label
                label_error = data["reference"].reference_phenotype.label_error
                fitness = data["reference"].reference_phenotype.fitness
                fitness_std = data["reference"].reference_phenotype.fitness_std

                yield BioCypherNode(
                    node_id=phenotype_id,
                    preferred_id=f"phenotype_{phenotype_id}",
                    node_label="phenotype",
                    properties={
                        "graph_level": graph_level,
                        "label": label,
                        "label_error": label_error,
                        "fitness": fitness,
                        "fitness_std": fitness_std,
                        "serialized_data": json.dumps(
                            data["reference"].reference_phenotype.model_dump()
                        ),
                    },
                )

    def _get_dataset_nodes(self) -> None:
        yield BioCypherNode(
            node_id="TmfKuzmin2018",
            preferred_id="TmfKuzmin2018",
            node_label="dataset",
        )

    def get_edges(self) -> None:
        logger.info("Generating edges.")
        logger.info("Get dataset experiment reference edges.")
        yield from self._get_dataset_experiment_ref_edges()
        logger.info("Get experiment dataset edges.")
        yield from self._get_experiment_dataset_edges()
        logger.info("Get experiment reference experiment edges.")
        yield from self._get_experiment_ref_experiment_edges()
        logger.info("Get genotype experiment edges.")
        logger.info("--- perturbation genotype edges.")
        yield from self._get_genotype_experiment_edges()
        logger.info("Get environment experiment edges.")
        yield from self._get_environment_experiment_edges()
        logger.info("Get environment experiment reference edges.")
        yield from self._get_environment_experiment_ref_edges()
        logger.info("Get phenotype experiment edges.")
        yield from self._get_phenotype_experiment_edges()
        logger.info("Get phenotype experiment reference edges.")
        yield from self._get_phenotype_experiment_ref_edges()
        logger.info("Get media environment edges.")
        yield from self._get_media_environment_edges()
        logger.info("Get temperature environment edges.")
        yield from self._get_temperature_environment_edges()
        logger.info("Get genome experiment reference edges.")
        yield from self._get_genome_edges()

    def _get_dataset_experiment_ref_edges(self):
        # concept level
        for data in tqdm(self.dataset.experiment_reference_index):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_ref_id,
                target_id="TmfKuzmin2018",
                relationship_label="experiment reference member of",
            )

    def _get_experiment_dataset_edges(self):
        # concept level
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            yield BioCypherEdge(
                source_id=experiment_id,
                target_id="TmfKuzmin2018",
                relationship_label="experiment member of",
            )

    def _get_experiment_ref_experiment_edges(self):
        # instance level
        print()
        for data in tqdm(self.dataset.experiment_reference_index):
            dataset_subset = self.dataset[data.index]
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            for i, data in enumerate(dataset_subset):
                experiment_id = hashlib.md5(
                    json.dumps(data["experiment"].model_dump()).encode("utf-8")
                ).hexdigest()
                yield BioCypherEdge(
                    source_id=experiment_ref_id,
                    target_id=experiment_id,
                    relationship_label="experiment reference of",
                )

    def _get_genotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        # CHECK if needed - don't think needed since exp ref index
        # seen_genotype_experiment_pairs: Set[tuple] = set()
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            for genotype in data["experiment"].genotype:
                genotype_id = hashlib.md5(
                    json.dumps(genotype.model_dump()).encode("utf-8")
                ).hexdigest()

                self._get_perturbation_genotype_edges(
                    genotype=genotype, genotype_id=genotype_id
                )

                # CHECK if needed - don't think needed since exp ref index
                # genotype_experiment_pair = (genotype_id, experiment_id)
                # if genotype_experiment_pair not in seen_genotype_experiment_pairs:
                #     seen_genotype_experiment_pairs.add(genotype_experiment_pair)

                yield BioCypherEdge(
                    source_id=genotype_id,
                    target_id=experiment_id,
                    relationship_label="genotype member of",
                )

    @staticmethod
    def _get_perturbation_genotype_edges(
        genotype: BaseGenotype, genotype_id: str
    ) -> Generator[BioCypherEdge, None, None]:
        if genotype.perturbation:
            perturbation_id = hashlib.md5(
                json.dumps(genotype.perturbation.model_dump()).encode("utf-8")
            ).hexdigest()

            yield BioCypherEdge(
                source_id=perturbation_id,
                target_id=genotype_id,
                relationship_label="perturbation member of",
            )

    def _get_environment_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_pairs: Set[tuple] = set()

        # Linking environments to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()

            env_experiment_pair = (environment_id, experiment_id)
            if env_experiment_pair not in seen_environment_experiment_pairs:
                seen_environment_experiment_pairs.add(env_experiment_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_id,
                    relationship_label="environment member of",
                )

    def _get_environment_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_environment_experiment_ref_pairs: Set[tuple] = set()

        # Linking environments to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            environment_id = hashlib.md5(
                json.dumps(data.reference.reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            env_experiment_ref_pair = (environment_id, experiment_ref_id)
            if env_experiment_ref_pair not in seen_environment_experiment_ref_pairs:
                seen_environment_experiment_ref_pairs.add(env_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_ref_id,
                    relationship_label="environment member of",
                )

    def _get_phenotype_experiment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_pairs: Set[tuple] = set()

        # Linking phenotypes to experiments
        for i, data in tqdm(enumerate(self.dataset)):
            experiment_id = hashlib.md5(
                json.dumps(data["experiment"].model_dump()).encode("utf-8")
            ).hexdigest()
            phenotype_id = hashlib.md5(
                json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
            ).hexdigest()

            phenotype_experiment_pair = (phenotype_id, experiment_id)
            if phenotype_experiment_pair not in seen_phenotype_experiment_pairs:
                seen_phenotype_experiment_pairs.add(phenotype_experiment_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_id,
                    relationship_label="phenotype member of",
                )

    def _get_phenotype_experiment_ref_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_phenotype_experiment_ref_pairs: Set[tuple] = set()

        # Linking phenotypes to experiment references
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            # Get the phenotype ID associated with the experiment reference
            phenotype_id = hashlib.md5(
                json.dumps(data.reference.reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            phenotype_experiment_ref_pair = (phenotype_id, experiment_ref_id)
            if phenotype_experiment_ref_pair not in seen_phenotype_experiment_ref_pairs:
                seen_phenotype_experiment_ref_pairs.add(phenotype_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=phenotype_id,
                    target_id=experiment_ref_id,
                    relationship_label="phenotype member of",
                )

    def _get_media_environment_edges(self) -> Generator[BioCypherEdge, None, None]:
        seen_media_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            media_id = hashlib.md5(
                json.dumps(data["experiment"].environment.media.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            media_environment_pair = (media_id, environment_id)
            if media_environment_pair not in seen_media_environment_pairs:
                seen_media_environment_pairs.add(media_environment_pair)

                yield BioCypherEdge(
                    source_id=media_id,
                    target_id=environment_id,
                    relationship_label="media member of",
                )

    def _get_temperature_environment_edges(
        self,
    ) -> Generator[BioCypherEdge, None, None]:
        seen_temperature_environment_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset)):
            environment_id = hashlib.md5(
                json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
            ).hexdigest()
            temperature_id = hashlib.md5(
                json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()

            temperature_environment_pair = (temperature_id, environment_id)
            if temperature_environment_pair not in seen_temperature_environment_pairs:
                seen_temperature_environment_pairs.add(temperature_environment_pair)

                yield BioCypherEdge(
                    source_id=temperature_id,
                    target_id=environment_id,
                    relationship_label="temperature member of",
                )

    def _get_genome_edges(self) -> None:
        seen_genome_experiment_ref_pairs: Set[tuple] = set()

        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.md5(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()

            genome_id = hashlib.md5(
                json.dumps(data.reference.reference_genome.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            genome_experiment_ref_pair = (genome_id, experiment_ref_id)
            if genome_experiment_ref_pair not in seen_genome_experiment_ref_pairs:
                seen_genome_experiment_ref_pairs.add(genome_experiment_ref_pair)

                yield BioCypherEdge(
                    source_id=genome_id,
                    target_id=experiment_ref_id,
                    relationship_label="genome member of",
                )


if __name__ == "__main__":
    ## Smf
    # Simple Testing
    # dataset = LmdbDatasetReader("data/torchcell/smf_kuzmin2018")
    # adapter = SmfKuzmin2018Adapter(dataset=dataset)
    # [i for i in adapter.get_nodes()]
    # [i for i in adapter.get_edges()]

    # # Advanced Testing
    # bc = BioCypher()
    # dataset = LmdbDatasetReader("data/torchcell/smf_kuzmin2018")
    # adapter = SmfKuzmin2018Adapter(dataset=dataset)
    # bc.write_nodes(adapter.get_nodes())
    # bc.write_edges(adapter.get_edges())
    # bc.write_import_call()
    # bc.write_schema_info(as_node=True)
    # bc.summary()

    ## Dmf
    # Simple Testing
    # dataset = LmdbDatasetReader("data/torchcell/dmf_kuzmin2018")
    # adapter = DmfKuzmin2018Adapter(dataset=dataset)
    # [i for i in adapter.get_nodes()]
    # [i for i in adapter.get_edges()]

    # # Advanced Testing
    # bc = BioCypher()
    # dataset = LmdbDatasetReader("data/torchcell/dmf_kuzmin2018")
    # adapter = DmfKuzmin2018Adapter(dataset=dataset)
    # bc.write_nodes(adapter.get_nodes())
    # bc.write_edges(adapter.get_edges())
    # bc.write_import_call()
    # bc.write_schema_info(as_node=True)
    # bc.summary()
    # print()

    ## Tmf
    # Simple Testing
    # dataset = LmdbDatasetReader("data/torchcell/tmf_kuzmin2018")
    # adapter = TmfKuzmin2018Adapter(dataset=dataset)
    # [i for i in adapter.get_nodes()]
    # [i for i in adapter.get_edges()]

    # # Advanced Testing
    bc = BioCypher()
    dataset = LmdbDatasetReader("data/torchcell/tmf_kuzmin2018")
    adapter = TmfKuzmin2018Adapter(dataset=dataset)
    bc.show_ontology_structure()
    bc.write_nodes(adapter.get_nodes())
    bc.write_edges(adapter.get_edges())
    bc.write_import_call()
    bc.write_schema_info(as_node=True)
    bc.summary()
    # print()



from .graph import (
    SCerevisiaeGraph,
    filter_by_contained_genes,
    filter_by_date,
    filter_go_IGI,
    filter_redundant_terms,
)

utils = [
    "filter_go_IGI",
    "filter_by_date",
    "filter_by_contained_genes",
    "filter_redundant_terms",
]

graphs = ["SCerevisiaeGraph"]

__all__ = utils + graphs



from abc import ABC, abstractmethod

import gffutils
import pandas as pd
from attrs import define, field
from Bio import SeqIO
from Bio.Seq import Seq


@define
class BaseGenome(ABC):
    @abstractmethod
    def get_sequence(self, chr: int, start: int, end: int) -> str:
        raise NotImplementedError

    # @abstractmethod
    # def get_chr_len(self, chr: int) -> int:
    #     # returns the length of the queried chromosome
    #     raise NotImplementedError

    @abstractmethod
    def get_gene_sequence(self, gene: str) -> str:
        raise NotImplementedError

    # @abstractmethod
    # def get_gene_position(self, gene: str) -> int:

    #     raise NotImplementedError
    @property
    @abstractmethod
    def translation_table(self) -> pd.DataFrame:
        raise NotImplementedError


def mismatch_positions(seq1, seq2):
    if len(seq1) != len(seq2):
        raise ValueError("Sequences must be the same length")
    mismatches = [i for i, (n1, n2) in enumerate(zip(seq1, seq2)) if n1 != n2]
    return mismatches


def get_chr_from_description(description: str) -> int:
    desc_split = [part.rstrip(",") for part in description.split()]
    if "mitochondrion" in desc_split:
        return 0
    else:
        roman_num = desc_split[desc_split.index("chromosome") + 1]
        return roman_to_int(roman_num)


def roman_to_int(s: str) -> int:
    roman_to_int_mapping = {
        "I": 1,
        "V": 5,
        "X": 10,
        "L": 50,
        "C": 100,
        "D": 500,
        "M": 1000,
    }
    result = 0
    for i in range(len(s)):
        if i > 0 and roman_to_int_mapping[s[i]] > roman_to_int_mapping[s[i - 1]]:
            result += roman_to_int_mapping[s[i]] - 2 * roman_to_int_mapping[s[i - 1]]
        else:
            result += roman_to_int_mapping[s[i]]
    return result


@define
class SCerevisiaeGenome(BaseGenome):
    _gff_path: str = (
        "data/ncbi/s_cerevisiae/ncbi_dataset/data/GCF_000146045.2/genomic.gff"
    )
    _fasta_path: str = "data/ncbi/s_cerevisiae/ncbi_dataset/data/GCF_000146045.2/GCF_000146045.2_R64_genomic.fna"
    db = field(init=False, repr=False)
    fasta_sequences = field(init=False, default=None, repr=False)
    chr_to_nc = field(init=False, default=None, repr=False)

    def __attrs_post_init__(self) -> None:
        # Create the database
        self.db = gffutils.create_db(
            self._gff_path,
            dbfn="data.db",
            force=True,
            keep_order=True,
            merge_strategy="merge",
            sort_attribute_values=True,
        )
        # Read the fasta file
        self.fasta_sequences = SeqIO.to_dict(SeqIO.parse(self._fasta_path, "fasta"))
        # Create mapping from chromosome number to sequence identifier
        self.chr_to_nc = {
            get_chr_from_description(self.fasta_sequences[key].description): key
            for key in self.fasta_sequences.keys()
        }

    def get_sequence(self, chr: int, start: int, end: int) -> str:
        # Get the identifier corresponding to the given chromosome number
        chr_id = self.chr_to_nc[chr]
        # Get the sequence for the chromosome with this identifier
        return str(self.fasta_sequences[chr_id].seq[start:end])

    def get_gene_sequence(self, gene_name: str) -> str:
        # Iterate through all genes in the gffutils database
        for gene_feature in self.db.features_of_type("gene"):
            # If the gene's Name attribute matches the provided gene_name
            if (
                "Name" in gene_feature.attributes
                and gene_feature["Name"][0] == gene_name
            ) or (
                "locus_tag" in gene_feature.attributes
                and gene_feature["locus_tag"][0] == gene_name
            ):
                # Extract the sequence for this gene from the fasta sequences
                sequence = self.fasta_sequences[gene_feature.seqid].seq[
                    gene_feature.start - 1 : gene_feature.end
                ]

                return str(sequence)

        # If no matching gene is found, raise an error
        raise ValueError(f"No gene found with name {gene_name}")

    @property
    def translation_table(self) -> pd.DataFrame:
        data = []
        for gene_feature in self.db.features_of_type("gene"):
            if (
                "locus_tag" in gene_feature.attributes
                and "Name" in gene_feature.attributes
            ):
                gene_id = gene_feature["locus_tag"][0]  # Get the locus_tag
                # Some of the locus_tag names do not match SGD Systematic Names.
                # This is especially true for mitochondrial genes. e.g. gene_id == "tE(UUC)Q"
                gene_name = gene_feature["Name"][0]  # Get the gene name
                data.append((gene_id, gene_name))
        return pd.DataFrame(data, columns=["locus_tag", "name"])


def main() -> None:
    genome = SCerevisiaeGenome()
    print(genome.get_sequence(1, 0, 10))  # Replace with valid parameters
    print(genome.get_gene_sequence("YFL039C"))  # Replace with valid gene
    print(genome.get_gene_sequence("ACT1"))  # Replace with valid gene
    assert genome.get_gene_sequence("YFL039C") == genome.get_gene_sequence(
        "ACT1"
    ), "genes should have same sequence"
    print(genome.translation_table)
    print()


if __name__ == "__main__":
    main()



# torchcell/datamodels/pydantic.py
# [[torchcell.datamodels.pydantic]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datamodels/pydantic.py
# Test file: torchcell/datamodels/test_pydantic.py
from pydantic import BaseModel


class ModelStrict(BaseModel):
    class Config:
        extra = "forbid"
        frozen = True


class ModelStrictArbitrary(BaseModel):
    class Config:
        extra = "forbid"
        frozen = True
        arbitrary_types_allowed = True



# torchcell/adapters/cell_adapter
# [[torchcell.adapters.cell_adapter]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/adapters/cell_adapter
# Test file: tests/torchcell/adapters/test_cell_adapter.py


from tqdm import tqdm
import hashlib
import json
from biocypher import BioCypher
from biocypher._create import BioCypherEdge, BioCypherNode
from biocypher._logger import get_logger
import logging
from typing import Generator, Set
import torch
from torchcell.datasets.scerevisiae import (
    SmfCostanzo2016Dataset,
    DmfCostanzo2016Dataset,
)
from torchcell.loader import CpuExperimentLoader
from torchcell.datamodels import Genotype
from concurrent.futures import ProcessPoolExecutor, as_completed
from torch_geometric.loader import DataLoader
from torchcell.dataset import Dataset
from concurrent.futures import ProcessPoolExecutor
import multiprocessing as mp
from functools import partial
from typing import Callable
from functools import wraps
from abc import abstractmethod
from omegaconf import OmegaConf
from typing import Any


class CellAdapter:
    def __init__(
        self,
        dataset: Dataset,
        compute_workers: int,
        io_workers: int,
        chunk_size: int = int(1e4),
        loader_batch_size: int = int(1e3),
    ):
        if loader_batch_size > chunk_size:
            raise ValueError(
                "chunk_size must be greater than or equal to loader_batch_size."
                "Our recommendation are chunk_size 2-3 order of magnitude in size."
            )
        self.dataset = dataset
        self.compute_workers = compute_workers
        self.io_workers = io_workers
        self.chunk_size = chunk_size
        self.loader_batch_size = loader_batch_size


    def get_data_by_type(self, chunk_processing_func: Callable):
        data_chunks = [
            self.dataset[i : i + self.chunk_size]
            for i in range(0, len(self.dataset), self.chunk_size)
        ]
        with ProcessPoolExecutor(max_workers=self.compute_workers) as executor:
            futures = [
                executor.submit(chunk_processing_func, chunk) for chunk in data_chunks
            ]
            for future in futures:
                for data in future.result():
                    yield data

    def data_chunker(data_creation_logic):
        @wraps(data_creation_logic)
        def decorator(self, data_chunk: dict):
            data_loader = CpuExperimentLoader(
                data_chunk,
                batch_size=self.loader_batch_size,
                num_workers=self.io_workers,
            )
            datas = []
            for batch in tqdm(data_loader):
                for data in batch:
                    transformed_data = data_chunk.transform_item(data)
                    data = data_creation_logic(self, transformed_data)
                    if isinstance(data, list):
                        datas.extend(data)  # Extend if 'data' is a list
                    else:
                        datas.append(data)  # Append if 'data' is a single item
            return datas

        return decorator

    # @abstractmethod
    # def get_nodes(self) -> Generator[BioCypherNode, None, None]:
    #     pass

    # @abstractmethod
    # def get_edges(self) -> Generator[BioCypherEdge, None, None]:
    #     pass

    # Nodes put here for now since we are trying to generalize as much as possible to reduce code here.

    def get_nodes(self):
        yield from self._get_experiment_reference_nodes()
        yield from self._get_genome_nodes()
        yield from self.get_data_by_type(self._experiment_node)
        yield from self.get_data_by_type(self._genotype_node)
        yield from self.get_data_by_type(self._perturbation_node)
        yield from self.get_data_by_type(self._environment_node)
        yield from self._get_reference_environment_nodes()
        yield from self.get_data_by_type(self._media_node)
        yield from self._get_reference_media_nodes()
        yield from self.get_data_by_type(self._temperature_node)
        yield from self._get_reference_temperature_nodes()
        yield from self.get_data_by_type(self._phenotype_node)
        yield from self._get_reference_phenotype_nodes()
        yield from self.get_dataset_nodes()

    def get_edges(self):
        yield from self.get_reference_dataset_edges()
        yield from self.get_data_by_type(self._experiment_dataset_edge)
        yield from self._get_reference_experiment_edges()
        yield from self.get_data_by_type(self._genotype_experiment_edge)
        yield from self.get_data_by_type(self._perturbation_genotype_edges)
        yield from self.get_data_by_type(self._environment_experiment_edges)
        yield from self._get_environment_experiment_ref_edges()
        yield from self.get_data_by_type(self._phenotype_experiment_edges)
        yield from self.get_data_by_type(self._media_environment_edge)
        yield from self.get_data_by_type(self._temperature_environment_edge)
        yield from self._get_genome_edges()

    # nodes
    def _get_experiment_reference_nodes(self) -> list[BioCypherNode]:
        nodes = []
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.sha256(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            node = BioCypherNode(
                node_id=experiment_ref_id,
                preferred_id="experiment reference",
                node_label="experiment reference",
                properties={"serialized_data": json.dumps(data.reference.model_dump())},
            )
            nodes.append(node)
        return nodes

    def _get_genome_nodes(self) -> list[BioCypherNode]:
        nodes = []
        seen_node_ids: Set[str] = set()
        for data in tqdm(self.dataset.experiment_reference_index):
            genome_id = hashlib.sha256(
                json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
            ).hexdigest()
            if genome_id not in seen_node_ids:
                seen_node_ids.add(genome_id)
                node = BioCypherNode(
                    node_id=genome_id,
                    preferred_id="genome",
                    node_label="genome",
                    properties={
                        "species": data.reference.reference_genome.species,
                        "strain": data.reference.reference_genome.strain,
                        "serialized_data": json.dumps(
                            data.reference.reference_genome.model_dump()
                        ),
                    },
                )
                nodes.append(node)
        return nodes

    @data_chunker
    def _experiment_node(self, data: dict) -> BioCypherNode:
        experiment_id = hashlib.sha256(
            json.dumps(data["experiment"].model_dump()).encode("utf-8")
        ).hexdigest()
        return BioCypherNode(
            node_id=experiment_id,
            preferred_id="experiment",
            node_label="experiment",
            properties={"serialized_data": json.dumps(data["experiment"].model_dump())},
        )

    @data_chunker
    def _genotype_node(self, data: dict) -> BioCypherNode:
        genotype = data["experiment"].genotype
        genotype_id = hashlib.sha256(
            json.dumps(genotype.model_dump()).encode("utf-8")
        ).hexdigest()
        return BioCypherNode(
            node_id=genotype_id,
            preferred_id="genotype",
            node_label="genotype",
            properties={
                "systematic_gene_names": genotype.systematic_gene_names,
                "perturbed_gene_names": genotype.perturbed_gene_names,
                "perturbation_types": genotype.perturbation_types,
                "serialized_data": json.dumps(genotype.model_dump()),
            },
        )

    @data_chunker
    def _perturbation_node(self, data: dict) -> BioCypherNode:
        perturbations = data["experiment"].genotype.perturbations
        for perturbation in perturbations:
            perturbation_id = hashlib.sha256(
                json.dumps(perturbation.model_dump()).encode("utf-8")
            ).hexdigest()
            return BioCypherNode(
                node_id=perturbation_id,
                preferred_id=perturbation.perturbation_type,
                node_label="perturbation",
                properties={
                    "systematic_gene_name": perturbation.systematic_gene_name,
                    "perturbed_gene_name": perturbation.perturbed_gene_name,
                    "perturbation_type": perturbation.perturbation_type,
                    "description": perturbation.description,
                    "strain_id": perturbation.strain_id,
                    "serialized_data": json.dumps(perturbation.model_dump()),
                },
            )

    @data_chunker
    def _environment_node(self, data: dict) -> BioCypherNode:
        environment_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
        ).hexdigest()
        media = json.dumps(data["experiment"].environment.media.model_dump())
        return BioCypherNode(
            node_id=environment_id,
            preferred_id="environment",
            node_label="environment",
            properties={
                "temperature": data["experiment"].environment.temperature.value,
                "media": media,
                "serialized_data": json.dumps(
                    data["experiment"].environment.model_dump()
                ),
            },
        )

    def _get_reference_environment_nodes(self) -> list[BioCypherNode]:
        nodes = []
        seen_node_ids = set()
        for data in tqdm(self.dataset.experiment_reference_index):
            environment_id = hashlib.sha256(
                json.dumps(data.reference.reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()
            if environment_id not in seen_node_ids:
                seen_node_ids.add(environment_id)
                media = json.dumps(
                    data.reference.reference_environment.media.model_dump()
                )
                node = BioCypherNode(
                    node_id=environment_id,
                    preferred_id="environment",
                    node_label="environment",
                    properties={
                        "temperature": data.reference.reference_environment.temperature.value,
                        "media": media,
                        "serialized_data": json.dumps(
                            data.reference.reference_environment.model_dump()
                        ),
                    },
                )
                nodes.append(node)
        return nodes

    @data_chunker
    def _media_node(self, data: dict) -> BioCypherNode:
        media_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.media.model_dump()).encode(
                "utf-8"
            )
        ).hexdigest()
        name = data["experiment"].environment.media.name
        state = data["experiment"].environment.media.state
        return BioCypherNode(
            node_id=media_id,
            preferred_id="media",
            node_label="media",
            properties={
                "name": name,
                "state": state,
                "serialized_data": json.dumps(
                    data["experiment"].environment.media.model_dump()
                ),
            },
        )

    def _get_reference_media_nodes(self) -> list[BioCypherNode]:
        seen_node_ids = set()
        nodes = []
        for data in tqdm(self.dataset.experiment_reference_index):
            media_id = hashlib.sha256(
                json.dumps(
                    data.reference.reference_environment.media.model_dump()
                ).encode("utf-8")
            ).hexdigest()
            if media_id not in seen_node_ids:
                seen_node_ids.add(media_id)
                name = data.reference.reference_environment.media.name
                state = data.reference.reference_environment.media.state
                node = BioCypherNode(
                    node_id=media_id,
                    preferred_id="media",
                    node_label="media",
                    properties={
                        "name": name,
                        "state": state,
                        "serialized_data": json.dumps(
                            data.reference.reference_environment.media.model_dump()
                        ),
                    },
                )
                nodes.append(node)
        return nodes

    @data_chunker
    def _temperature_node(self, data: dict) -> BioCypherNode:
        temperature_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.temperature.model_dump()).encode(
                "utf-8"
            )
        ).hexdigest()
        return BioCypherNode(
            node_id=temperature_id,
            preferred_id="temperature",
            node_label="temperature",
            properties={
                "value": data["experiment"].environment.temperature.value,
                "unit": data["experiment"].environment.temperature.unit,
                "serialized_data": json.dumps(
                    data["experiment"].environment.temperature.model_dump()
                ),
            },
        )

    def _get_reference_temperature_nodes(self) -> list[BioCypherNode]:
        nodes = []
        seen_node_ids: Set[str] = set()
        for data in tqdm(self.dataset.experiment_reference_index):
            temperature_id = hashlib.sha256(
                json.dumps(
                    data.reference.reference_environment.temperature.model_dump()
                ).encode("utf-8")
            ).hexdigest()
            if temperature_id not in seen_node_ids:
                seen_node_ids.add(temperature_id)
                node = BioCypherNode(
                    node_id=temperature_id,
                    preferred_id="temperature",
                    node_label="temperature",
                    properties={
                        "value": data.reference.reference_environment.temperature.value,
                        "unit": data.reference.reference_environment.temperature.unit,
                        "serialized_data": json.dumps(
                            data.reference.reference_environment.temperature.model_dump()
                        ),
                    },
                )
                nodes.append(node)
        return nodes

    @data_chunker
    def _phenotype_node(self, data: dict) -> BioCypherNode:
        phenotype_id = hashlib.sha256(
            json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
        ).hexdigest()

        graph_level = data["experiment"].phenotype.graph_level
        label = data["experiment"].phenotype.label
        label_error = data["experiment"].phenotype.label_error
        fitness = data["experiment"].phenotype.fitness
        fitness_std = data["experiment"].phenotype.fitness_std

        return BioCypherNode(
            node_id=phenotype_id,
            preferred_id=f"phenotype_{phenotype_id}",
            node_label="phenotype",
            properties={
                "graph_level": graph_level,
                "label": label,
                "label_error": label_error,
                "fitness": fitness,
                "fitness_std": fitness_std,
                "serialized_data": json.dumps(
                    data["experiment"].phenotype.model_dump()
                ),
            },
        )

    def _get_reference_phenotype_nodes(self) -> list[BioCypherNode]:
        nodes = []
        for data in tqdm(self.dataset.experiment_reference_index):
            phenotype_id = hashlib.sha256(
                json.dumps(data.reference.reference_phenotype.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()

            graph_level = data.reference.reference_phenotype.graph_level
            label = data.reference.reference_phenotype.label
            label_error = data.reference.reference_phenotype.label_error
            fitness = data.reference.reference_phenotype.fitness
            fitness_std = data.reference.reference_phenotype.fitness_std

            node = BioCypherNode(
                node_id=phenotype_id,
                preferred_id=f"phenotype",
                node_label="phenotype",
                properties={
                    "graph_level": graph_level,
                    "label": label,
                    "label_error": label_error,
                    "fitness": fitness,
                    "fitness_std": fitness_std,
                    "serialized_data": json.dumps(
                        data.reference.reference_phenotype.model_dump()
                    ),
                },
            )
            nodes.append(node)
        return nodes

    def get_dataset_nodes(self) -> list[BioCypherNode]:
        nodes = [
            BioCypherNode(
                node_id=self.dataset.__class__.__name__,
                preferred_id=self.dataset.__class__.__name__,
                node_label="dataset",
            )
        ]
        return nodes

    # edges
    def get_reference_dataset_edges(self) -> list[BioCypherEdge]:
        edges = []
        for data in self.dataset.experiment_reference_index:
            reference_id = hashlib.sha256(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            edge = BioCypherEdge(
                source_id=reference_id,
                target_id=self.dataset.__class__.__name__,
                relationship_label="experiment reference member of",
            )
            edges.append(edge)
        return edges

    @data_chunker
    def _experiment_dataset_edge(self, data: dict) -> list[BioCypherEdge]:
        experiment_id = hashlib.sha256(
            json.dumps(data["experiment"].model_dump()).encode("utf-8")
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=experiment_id,
            target_id=self.dataset.__class__.__name__,
            relationship_label="experiment member of",
        )
        return edge

    def _get_reference_experiment_edges(self) -> list[BioCypherEdge]:
        edges = []
        for data in tqdm(self.dataset.experiment_reference_index):
            dataset_subset = self.dataset[torch.tensor(data.index)]
            experiment_ref_id = hashlib.sha256(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            for i, data in enumerate(dataset_subset):
                data = self.dataset.transform_item(data)
                experiment_id = hashlib.sha256(
                    json.dumps(data["experiment"].model_dump()).encode("utf-8")
                ).hexdigest()
                edge = BioCypherEdge(
                    source_id=experiment_ref_id,
                    target_id=experiment_id,
                    relationship_label="experiment reference of",
                )
                edges.append(edge)
        return edges

    @data_chunker
    def _genotype_experiment_edge(self, data: dict) -> BioCypherEdge:
        experiment_id = hashlib.sha256(
            json.dumps(data["experiment"].model_dump()).encode("utf-8")
        ).hexdigest()
        genotype = data["experiment"].genotype
        genotype_id = hashlib.sha256(
            json.dumps(genotype.model_dump()).encode("utf-8")
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=genotype_id,
            target_id=experiment_id,
            relationship_label="genotype member of",
        )
        return edge

    @data_chunker
    def _perturbation_genotype_edges(self, data: dict) -> list[BioCypherEdge]:
        edges = []
        genotype = data["experiment"].genotype
        for perturbation in genotype.perturbations:
            genotype_id = hashlib.sha256(
                json.dumps(genotype.model_dump()).encode("utf-8")
            ).hexdigest()
            perturbation_id = hashlib.sha256(
                json.dumps(perturbation.model_dump()).encode("utf-8")
            ).hexdigest()
            edges.append(
                BioCypherEdge(
                    source_id=perturbation_id,
                    target_id=genotype_id,
                    relationship_label="perturbation member of",
                )
            )
            return edges

    @data_chunker
    def _environment_experiment_edges(self, data: dict) -> BioCypherEdge:
        experiment_id = hashlib.sha256(
            json.dumps(data["experiment"].model_dump()).encode("utf-8")
        ).hexdigest()
        environment_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=environment_id,
            target_id=experiment_id,
            relationship_label="environment member of",
        )
        return edge

    def _get_environment_experiment_ref_edges(self) -> list[BioCypherEdge]:
        edges = []
        seen_environment_experiment_ref_pairs: Set[tuple] = set()
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.sha256(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            environment_id = hashlib.sha256(
                json.dumps(data.reference.reference_environment.model_dump()).encode(
                    "utf-8"
                )
            ).hexdigest()
            env_experiment_ref_pair = (environment_id, experiment_ref_id)
            if env_experiment_ref_pair not in seen_environment_experiment_ref_pairs:
                seen_environment_experiment_ref_pairs.add(env_experiment_ref_pair)

                edge = BioCypherEdge(
                    source_id=environment_id,
                    target_id=experiment_ref_id,
                    relationship_label="environment member of",
                )
                edges.append(edge)
        return edges

    @data_chunker
    def _phenotype_experiment_edges(self, data: dict) -> BioCypherEdge:
        experiment_id = hashlib.sha256(
            json.dumps(data["experiment"].model_dump()).encode("utf-8")
        ).hexdigest()
        phenotype_id = hashlib.sha256(
            json.dumps(data["experiment"].phenotype.model_dump()).encode("utf-8")
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=phenotype_id,
            target_id=experiment_id,
            relationship_label="phenotype member of",
        )
        return edge

    @data_chunker
    def _media_environment_edge(self, data: dict) -> BioCypherEdge:
        environment_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
        ).hexdigest()
        media_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.media.model_dump()).encode(
                "utf-8"
            )
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=media_id,
            target_id=environment_id,
            relationship_label="media member of",
        )
        return edge

    @data_chunker
    def _temperature_environment_edge(self, data: dict) -> BioCypherEdge:
        environment_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.model_dump()).encode("utf-8")
        ).hexdigest()
        temperature_id = hashlib.sha256(
            json.dumps(data["experiment"].environment.temperature.model_dump()).encode(
                "utf-8"
            )
        ).hexdigest()
        edge = BioCypherEdge(
            source_id=temperature_id,
            target_id=environment_id,
            relationship_label="temperature member of",
        )
        return edge

    def _get_genome_edges(self) -> list[BioCypherEdge]:
        edges = []
        seen_genome_experiment_ref_pairs: Set[tuple] = set()
        for i, data in tqdm(enumerate(self.dataset.experiment_reference_index)):
            experiment_ref_id = hashlib.sha256(
                json.dumps(data.reference.model_dump()).encode("utf-8")
            ).hexdigest()
            genome_id = hashlib.sha256(
                json.dumps(data.reference.reference_genome.model_dump()).encode("utf-8")
            ).hexdigest()
            genome_experiment_ref_pair = (genome_id, experiment_ref_id)
            if genome_experiment_ref_pair not in seen_genome_experiment_ref_pairs:
                seen_genome_experiment_ref_pairs.add(genome_experiment_ref_pair)
                edge = BioCypherEdge(
                    source_id=genome_id,
                    target_id=experiment_ref_id,
                    relationship_label="genome member of",
                )
                edges.append(edge)
        return edges


if __name__ == "__main__":
    pass



import requests
import os
import zipfile

url = "https://api.ncbi.nlm.nih.gov/datasets/v2alpha/genome/accession/GCF_000146045.2/download"

params = {
    "include_annotation_type": "GENOME_FASTA,GENOME_GFF,RNA_FASTA,CDS_FASTA,PROT_FASTA,SEQUENCE_REPORT",
    "filename": "GCF_000146045.2.zip",
}

headers = {"Accept": "application/zip"}

response = requests.get(url, headers=headers, params=params, stream=True)

# Specify the directory where you want to save the file
directory = "data/ncbi/s_cerevisiae/"

# Create the directory if it doesn't exist
os.makedirs(directory, exist_ok=True)

# Now the file will be saved to the specified directory
filename = directory + params["filename"]

with open(filename, "wb") as d:
    for chunk in response.iter_content(chunk_size=128):
        d.write(chunk)

# Unzip the file in the same location
with zipfile.ZipFile(filename, "r") as zip_ref:
    zip_ref.extractall(directory)

# Delete the .zip file
os.remove(filename)



# db_cli.py
from cliff.app import App
from cliff.commandmanager import CommandManager
from torchcell.database import BuildCommand  # Import the command


class TCDB(App):

    def __init__(self):
        super(TCDB, self).__init__(
            description="Database CLI",
            version="0.1",
            command_manager=CommandManager("db.cli"),
            deferred_help=True,
        )
        self.command_manager.add_command("build", BuildCommand)


def main(argv=None):
    myapp = TCDB()
    return myapp.run(argv)


if __name__ == "__main__":
    import sys

    sys.exit(main(sys.argv[1:]))



from .dcell import DCellLoss
from .weighted_mse import WeightedMSELoss

standard_losses = {"weighted_mse": WeightedMSELoss}

model_losses = {"dcell": DCellLoss}

__all__ = ["standard_losses", "model_losses"]



import os

import torch
from transformers import AutoModelForMaskedLM, AutoTokenizer

from torchcell.models.llm import NucleotideModel

MODEL_NAME = "InstaDeepAI/nucleotide-transformer-2.5b-multi-species"


class NucleotideTransformer(NucleotideModel):
    def __init__(self):
        self.tokenizer = None
        self.model = None
        self.load_model()

    @staticmethod
    def _check_and_download_model():
        # Define the model name

        # Define the directory where you want the model to be saved
        script_dir = os.path.dirname(os.path.realpath(__file__))
        target_directory = os.path.join(
            script_dir, "pretrained_LLM", "nucleotide_transformer"
        )

        # Create the target directory if it doesn't exist
        if not os.path.exists(target_directory):
            os.makedirs(target_directory)

        model_directory = os.path.join(target_directory, MODEL_NAME)

        # Check if the model has already been downloaded
        if os.path.exists(model_directory):
            print(f"{MODEL_NAME} model already downloaded.")
        else:
            print(f"Downloading {MODEL_NAME} model to {model_directory}...")
            # tokenizer
            AutoTokenizer.from_pretrained(MODEL_NAME, cache_dir=target_directory)
            # model
            AutoModelForMaskedLM.from_pretrained(MODEL_NAME, cache_dir=target_directory)
            print("Download finished.")

    @property
    def max_sequence_size(self):
        # Found empirically... listed as 6kb in paper.
        return 5979

    def load_model(self):
        # Check and download the model if necessary
        self._check_and_download_model()

        # Load the tokenizer and the model
        self.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
        self.model = AutoModelForMaskedLM.from_pretrained(MODEL_NAME)

    def embed(self, sequences: list[str], mean_embedding: bool = False) -> torch.Tensor:
        tokens_ids = self.tokenizer.batch_encode_plus(sequences, return_tensors="pt")[
            "input_ids"
        ]
        # TODO check this error handling... looks wrong with 1000
        for token_id, sequence in zip(tokens_ids, sequences):
            if len(token_id) > 1000:
                raise ValueError(
                    f"Sequence length ({len(sequence)}) is too long for this model. "
                    f"Max sequence size is {self.max_size}."
                    f"Token length is {len(token_id)}."
                    f"Max token length is 1000."
                )
        # Compute the embeddings
        attention_mask = tokens_ids != self.tokenizer.pad_token_id
        torch_outs = self.model(
            tokens_ids,
            attention_mask=attention_mask,
            encoder_attention_mask=attention_mask,
            output_hidden_states=True,
        )

        embeddings = torch_outs["hidden_states"][-1].detach()

        if mean_embedding:
            # Compute mean embeddings per sequence
            embeddings = torch.sum(
                attention_mask.unsqueeze(-1) * embeddings, axis=-2
            ) / torch.sum(attention_mask, axis=-1).unsqueeze(-1)

        return embeddings


def main():
    # Initialize the NucleotideTransformer
    transformer = NucleotideTransformer()

    # Create a dummy dna sequence
    sequences = ["ACTACG" * 100, "ACTACG" * 100]
    sequences = [i[:5979] for i in sequences]
    print(len(sequences[0]))
    # Get embeddings
    embeddings = transformer.embed(sequences)

    # print(f"Embeddings per token: {embeddings}")
    print(f"Embeddings shape: {embeddings.shape}")

    # Get Mean embeddings
    embeddings = transformer.embed(sequences, mean_embedding=True)

    # print(f"Embeddings per token: {embeddings}")
    print(f"Embeddings shape: {embeddings.shape}")


if __name__ == "__main__":
    main()



import os
import os.path as osp
from abc import ABC, abstractmethod
from collections import Counter
from datetime import datetime

import matplotlib.pyplot as plt
from attrs import define, field
from sklearn import base

from torchcell.sequence import Genome
from torchcell.sgd.sequence import SCerevisiaeGenome
import torchcell

style_file_path = osp.join(osp.dirname(torchcell.__file__), 'torchcell.mplstyle')
plt.style.use(style_file_path)

class GenomePlot(ABC):
    genome: Genome

    @abstractmethod
    def plot() -> None:
        raise NotImplementedError("Subclasses must implement plot() method.")

    def save(self, as_pdf: bool = False, as_pickle: bool = False) -> None:
        current_dir = osp.dirname(osp.abspath(__file__))
        # Get current date and time and format it as year-month-day-hour-minute-second.
        timestamp = datetime.datetime.now().strftime("%Y.%m.%d.%H.%M.%S")
        file_path = osp.join(
            ".notes/assets/images",
            osp.relpath(current_dir),
            f"{self.__class__.__name__}-{timestamp}",  # Append timestamp to file name.
        )
        dir_name = osp.dirname(file_path)
        os.makedirs(dir_name, exist_ok=True)
        plt.savefig(f"{file_path}.png", format="png", bbox_inches="tight")
        print(f"Saved figure: {file_path}.png")
        if as_pdf:
            plt.savefig(f"{file_path}.pdf", format="pdf", bbox_inches="tight")
            print(f"Saved figure: {file_path}.pdf")
        if as_pickle:
            with open(f"{file_path}.pkl", "wb") as f:
                pickle.dump(plt.gcf(), f)  # Save the figure using pickle
            print(f"Saved figure data: {file_path}.pkl")


@define
class PlotFeatureTypeCounts(GenomePlot):
    genome: Genome
    plt: plt = field(init=False)

    @property
    def feature_type_counts(self):
        feature_types = [
            feat.featuretype for feat in list(self.genome.db.all_features())
        ]
        return dict(Counter(feature_types))

    def plot(self) -> None:
        feature_type_counts = self.feature_type_counts
        types = list(feature_type_counts.keys())
        counts = list(feature_type_counts.values())

        # Set the figure size based on the number of feature types.
        plt.figure(figsize=(10, len(types) * 0.4))

        bars = plt.barh(types, counts)
        plt.xlabel("Count")
        plt.ylabel("Feature type")
        plt.title("Feature Type Counts")

        # Add counts as text on the bars.
        for bar in bars:
            width = bar.get_width()
            plt.text(
                width,
                bar.get_y() + bar.get_height() / 2,
                f" {width}",
                va="center",
                ha="left",  # Align the labels to the left
            )

        # Adjust xlim to provide more space for the labels
        plt.xlim(0, max(counts) * 1.1)  # Increase xlim by 10%

        self.plt = plt

    def show(self):
        plt.show()

    def close(self):
        plt.close()


def main():
    genome = SCerevisiaeGenome()
    genome_plot = PlotFeatureTypeCounts(genome)
    genome_plot.plot()
    genome_plot.save(pdf=True, mpl=True)


if __name__ == "__main__":
    main()



# torchcell/models/llm.py
# [[torchcell.models.llm]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/llm.py
# Test file: torchcell/models/test_llm.py

from abc import ABC, abstractmethod

import torch
from attrs import define
from transformers import AutoModelForMaskedLM, AutoTokenizer


class NucleotideModel(ABC):
    def __init__(self, model_name: str):
        self.tokenizer = None
        self.model = None
        self.load_model(model_name)

    @property
    def max_sequence_size(self) -> int:
        """Returns the maximum sequence size for the transformer model."""
        if self._max_sequence_size is None:
            raise ValueError("Max size has not been set for this model.")
        return self._max_sequence_size

    @staticmethod
    @abstractmethod
    def _check_and_download_model():
        pass

    @abstractmethod
    def load_model(self, model_name: str):
        pass

    @abstractmethod
    def embed(self, sequences: list, mean_embedding: bool = False) -> torch.Tensor:
        pass


class PeptideModel(ABC):
    def __init__(self, model_name: str):
        self.tokenizer = None
        self.model = None
        self.load_model(model_name)

    @property
    def max_sequence_size(self) -> int:
        """Returns the maximum sequence size for the transformer model."""
        if self._max_sequence_size is None:
            raise ValueError("Max size has not been set for this model.")
        return self._max_sequence_size

    @staticmethod
    @abstractmethod
    def _check_and_download_model(model_name: str):
        pass

    @abstractmethod
    def load_model(self, model_name: str):
        pass

    @abstractmethod
    def embed(self, sequences: list, mean_embedding: bool = False) -> torch.Tensor:
        pass


@define
class pretrained_LLM:
    tokenizer: AutoTokenizer
    model: AutoModelForMaskedLM


if __name__ == "__main__":
    pass



import os
import os.path as osp

from torchcell.datasets import (
    CodonFrequencyDataset,
    FungalUpDownTransformerDataset,
    OneHotGeneDataset,
    ProtT5Dataset,
)
from torchcell.sequence.genome.scerevisiae.s288c import SCerevisiaeGenome

if __name__ == "__main__":
    from dotenv import load_dotenv

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(data_root=osp.join(DATA_ROOT, "data/sgd/genome"))

    fud_downstream = FungalUpDownTransformerDataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/fungal_up_down_embed"),
        genome=genome,
        model_name="species_downstream",
    )

    fud_upstream = FungalUpDownTransformerDataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/fungal_up_down_embed"),
        genome=genome,
        model_name="species_upstream",
    )

    codon_freq = CodonFrequencyDataset(
        root="data/scerevisiae/codon_frequency", genome=genome
    )

    one_hot_gene = OneHotGeneDataset(
        root="data/scerevisiae/gene_one_hot", genome=genome
    )
    prot_T5_dataset_no_dubious = ProtT5Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/protT5_embed"),
        genome=genome,
        model_name="prot_t5_xl_uniref50_no_dubious",
    )
    prot_T5_dataset_all = ProtT5Dataset(
        root=osp.join(DATA_ROOT, "data/scerevisiae/protT5_embed"),
        genome=genome,
        model_name="prot_t5_xl_uniref50_all",
    )
    print()
    # print(fud_downstream)
    # print(fud_downstream[100])
    # print(one_hot_gene)
    # print(one_hot_gene[100])
    # # BUG if the datasets are not the same size they will not be added properly. Should make sure that all datasets are the same size. Add.
    # dataset = fud_downstream + one_hot_gene
    # print(dataset)
    # print(dataset[100])



from .data import (
    ExperimentReferenceIndex,
    ReferenceIndex,
    serialize_for_hashing,
    compute_sha256_hash,
)

data = [
    "ExperimentReferenceIndex",
    "ReferenceIndex",
    "serialize_for_hashing",
    "compute_md5_hash",
]


__all__ = data



#############
from intermine.webservice import Service


# def ym_kb_plus(gene="rad54"):
#     service = Service("https://yeastmine.yeastgenome.org/yeastmine/service")
#     template = service.get_template("Gene_Flanking_Sequence")

#     # You can edit the constraint values below
#     # B    Gene
#     # C    Gene.flankingRegions.direction
#     # A    Gene.flankingRegions.distance
#     # D    Gene.flankingRegions.includeGene

#     rows = template.rows(
#         B={"op": "LOOKUP", "value": gene, "extra_value": "S. cerevisiae"},
#         C={"op": "=", "value": "both"},
#         A={"op": "=", "value": "1.0kb"},
#         D={"op": "=", "value": "true"},
#     )
#     for row in rows:
#         print(
#             row["secondaryIdentifier"],
#             row["symbol"],
#             row["length"],
#             row["flankingRegions.direction"],
#             row["flankingRegions.sequence.length"],
#             row["flankingRegions.sequence.residues"],
#         )
#         print()

# from Bio import Entrez, SeqIO


# def download_genome(accession_number, filename):
#     Entrez.email = "michaeljvolk7@gmail.com"  # Always tell NCBI who you are
#     # First, we need to use esearch to find the IDs of the sequences we're interested in
#     handle = Entrez.esearch(db="nucleotide", term=accession_number, retmax=10000)
#     record = Entrez.read(handle)
#     id_list = record["IdList"]
#     handle.close()
#     # Then, we can fetch the sequences
#     handle = Entrez.efetch(db="nucleotide", id=id_list, rettype="gb", retmode="text")
#     # Save to a file
#     with open(filename, "w") as out_handle:
#         out_handle.write(handle.read())
#     handle.close()


# def load_genome(filename):
#     with open(filename, "r") as handle:
#         return list(SeqIO.parse(handle, "genbank"))


# def extract_genes(genomes, upstream_length, downstream_length):
#     for genome in genomes:
#         for feature in genome.features:
#             if feature.type == "CDS":
#                 location = feature.location
#                 start = location.start
#                 end = location.end
#                 extended_start = max(0, start - upstream_length)
#                 extended_end = min(len(genome.seq), end + downstream_length)
#                 # Extract the gene sequence
#                 gene_seq = genome.seq[extended_start:extended_end]
#                 # Yield the gene sequence and its location
#                 yield gene_seq, location


# def main():
#     genome_file = "yeast_genome.gb"
#     download_genome("GCF_000146045.2", genome_file)
#     yeast_genomes = load_genome(genome_file)
#     upstream_length = 100  # Adjust as needed
#     downstream_length = 100  # Adjust as needed
#     genes = []
#     for gene_seq, location in extract_genes(
#         yeast_genomes, upstream_length, downstream_length
#     ):
#         genes.append(str(gene_seq))  # Convert Seq to str
#     print(genes)  # Prints the sequences of genes

from Bio import Entrez, SeqIO


def download_gene(gene_name, filename):
    Entrez.email = "michaeljvolk7@gmail.com"  # Always tell NCBI who you are
    handle = Entrez.esearch(
        db="nucleotide", term=f"{gene_name} Saccharomyces cerevisiae[Orgn]", retmax=1
    )
    record = Entrez.read(handle)
    id_list = record["IdList"]
    handle.close()

    if len(id_list) > 0:  # if a gene is found
        handle = Entrez.efetch(
            db="nucleotide", id=id_list[0], rettype="gb", retmode="text"
        )
        with open(filename, "w") as out_handle:
            out_handle.write(handle.read())
        handle.close()
    else:
        print(f"No gene found for {gene_name}")


def main():
    gene_file = "gene.gb"
    download_gene("YOR202W", gene_file)


if __name__ == "__main__":
    # ym_kb_plus()
    # other()
    # another()
    main()



# torchcell/models/act.py
# [[torchcell.models.act]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/act.py
# Test file: torchcell/models/test_act.py

from torch import nn

act_register = {
    "relu": nn.ReLU(),
    "gelu": nn.GELU(),
    "sigmoid": nn.Sigmoid(),
    "leaky_relu": nn.LeakyReLU(),  # TODO add params, look at graph gym
    "tanh": nn.Tanh(),
}



from .kuzmin2018_pypy_adapter import (
    SmfKuzmin2018Adapter,
    DmfKuzmin2018Adapter,
    TmfKuzmin2018Adapter,
)
from .costanzo2016_pypy_adapter import SmfCostanzo2016Adapter, DmfCostanzo2016Adapter


kuzmin2018_adapters = [
    "SmfKuzmin2018Adapter",
    "DmfKuzmin2018Adapter",
    "TmfKuzmin2018Adapter",
]
costanzo2016_adapters = ["SmfCostanzo2016Adapter", "DmfCostanzo2016Adapter"]

__all__ = kuzmin2018_adapters + costanzo2016_adapters



# torchcell/datasets/__init__.py
# [[torchcell.datasets.__init__]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datasets/__init__.py
# Test file: tests/torchcell/datasets/test___init__.py
from .cell import CellDataset
from .codon_frequency import CodonFrequencyDataset
from .dcell import DCellDataset

# from .costanzo2016 import DmfCostanzo2016Dataset, SmfCostanzo2016Dataset # in scerevisiae
from .fungal_up_down_transformer import FungalUpDownTransformerDataset
from .nucleotide_transformer import NucleotideTransformerDataset
from .one_hot_gene import OneHotGeneDataset
from .protT5 import ProtT5Dataset

core_datasets = ["CellDataset", "DCellDataset"]

embedding_datasets = [
    "NucleotideTransformerDataset",
    "FungalUpDownTransformerDataset",
    "CodonFrequencyDataset",
    "OneHotGeneDataset",
    "ProtT5Dataset",
]


# experiment_datasets = ["SmfCostanzo2016Dataset", "DmfCostanzo2016Dataset"]


# + experiment_datasets

__all__ = core_datasets + embedding_datasets



from .dcell_regression import DCellRegressionTask
from .dcell_regression_slim import DCellRegressionSlimTask
from .graph_convolution_regression import GraphConvRegressionTask
from .regression import RegressionTask
from .simple_linear_regression import SimpleLinearRegressionTask

__all__ = [
    "RegressionTask",
    "SimpleLinearRegressionTask",
    "GraphConvRegressionTask",
    "DCellRegressionTask",
]



# @define
# class BaseGenome(ABC):
#     @abstractmethod
#     def get_sequence(self, chr: int, start: int, end: int) -> str:
#         raise NotImplementedError

#     # @abstractmethod
#     # def get_chr_len(self, chr: int) -> int:
#     #     # returns the length of the queried chromosome
#     #     raise NotImplementedError

#     @abstractmethod
#     def get_gene_sequence(self, gene: str) -> str:
#         raise NotImplementedError

#     # @abstractmethod
#     # def get_gene_position(self, gene: str) -> int:

#     #     raise NotImplementedError
#     @property
#     @abstractmethod
#     def translation_table(self) -> pd.DataFrame:
#         raise NotImplementedError

#     # @abstractmethod
#     # __getitem__(self, key: str) -> str:



# torchcell/datamodules/cell.py
# [[torchcell.datamodules.cell]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/datamodules/cell.py
# Test file: torchcell/datamodules/test_cell.py

import lightning as L
import torch
from torch_geometric.loader import DataLoader


class CellDataModule(L.LightningDataModule):
    def __init__(
        self,
        dataset,
        batch_size: int = 32,
        num_workers: int = 0,
        pin_memory: bool = False,
    ):
        super().__init__()
        self.dataset = dataset
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.pin_memory = pin_memory
        self.train_epoch_size = None
        self.train_ratio = 0.8
        self.val_ratio = 0.1
        self.train_epoch_size = int(
            len(self.dataset) * self.train_ratio / self.batch_size
        )

    def setup(self, stage=None):
        # Split the dataset into train, val, and test sets
        num_train = int(self.train_ratio * len(self.dataset))
        num_val = int(self.val_ratio * len(self.dataset))
        num_test = len(self.dataset) - num_train - num_val

        (
            self.train_dataset,
            self.val_dataset,
            self.test_dataset,
        ) = torch.utils.data.random_split(self.dataset, [num_train, num_val, num_test])

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            follow_batch=["x"]
            # follow_batch=["x", "x_pert", "x_one_hop_pert"],
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            follow_batch=["x"]
            # follow_batch=["x", "x_pert", "x_one_hop_pert"],
        )

    def test_dataloader(self):
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
            follow_batch=["x"]
            # follow_batch=["x", "x_pert", "x_one_hop_pert"],
        )


if __name__ == "__main__":
    pass



from .costanzo2016_adapter import SmfCostanzo2016Adapter, DmfCostanzo2016Adapter
from .kuzmin2018_adapter import (
    SmfKuzmin2018Adapter,
    DmfKuzmin2018Adapter,
    TmfKuzmin2018Adapter,
)
from .cell_adapter import CellAdapter

cell_adapters = ["CellAdapter"]

# Experiment Adapters

# Dataset Adapters
costanzo2016_adapters = ["SmfCostanzo2016Adapter", "DmfCostanzo2016Adapter"]

kuzmin2018_adapters = [
    "SmfKuzmin2018Adapter",
    "DmfKuzmin2018Adapter",
    "TmfKuzmin2018Adapter",
]

__all__ = cell_adapters + costanzo2016_adapters + kuzmin2018_adapters



# torchcell/data/data.py
# [[torchcell.data.data]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/data/data.py
# Test file: tests/torchcell/data/test_data.py

import hashlib
import json
from typing import List

from pydantic import field_validator

from torchcell.datamodels import (
    ExperimentReference,
    FitnessExperimentReference,
    ModelStrict,
)

from typing import Union

# List of classes that can be part of ExperimentReference
ExperimentReferenceType = ExperimentReference | FitnessExperimentReference


class ExperimentReferenceIndex(ModelStrict):
    reference: ExperimentReferenceType
    index: List[bool]

    def __repr__(self):
        if len(self.index) > 5:
            return f"ExperimentReferenceIndex(reference={self.reference}, index={self.index[:5]}...)"
        else:
            return f"ExperimentReferenceIndex(reference={self.reference}, index={self.index})"


class ReferenceIndex(ModelStrict):
    data: List[ExperimentReferenceIndex]

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return len(self.data)

    def __iter__(self):
        return iter(self.data)

    @field_validator("data")
    def validate_data(cls, v):
        summed_indices = sum(
            [
                boolean_value
                for exp_ref_index in v
                for boolean_value in exp_ref_index.index
            ]
        )

        if summed_indices != len(v[0].index):
            raise ValueError("Sum of indices must equal the number of experiments")
        return v


# def serialize_for_hashing(obj) -> str:
#     """
#     Serialize a Pydantic object for hashing.
#     """
#     return json.dumps(obj.dict(), sort_keys=True)


# HACK temporary not using pydantic return on get item
def serialize_for_hashing(obj) -> str:
    """
    Serialize a Pydantic object for hashing.
    """
    return json.dumps(obj, sort_keys=True)


def compute_sha256_hash(content: str) -> str:
    """
    Compute the sha256 hash of a string.
    """
    return hashlib.sha256(content.encode()).hexdigest()



from goatools.obo_parser import GODag

# Initialize the GO ontology DAG (Directed Acyclic Graph)
go_dag = GODag("data/go/go.obo")

# GO Term to be checked
go_term = "GO:0000185"

# Check if the term exists in the ontology
if go_term in go_dag:
    # Check if it is obsolete
    if go_dag[go_term].is_obsolete:
        print(f"{go_term} is deprecated.")
    else:
        print(f"{go_term} is not deprecated.")
else:
    print(f"{go_term} does not exist in the loaded ontology.")

print()



# torchcell/sequence/genome/scerevisiae/s288c_gb.py
# [[torchcell.sequence.genome.scerevisiae.s288c_gb]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/sequence/genome/scerevisiae/s288c_gb.py
# Test file: torchcell/sequence/genome/scerevisiae/test_s288c_gb.py

import glob
import gzip
import logging
import os
import os.path as osp
import shutil
import subprocess
import tarfile
import zipfile
from typing import Set

import gffutils
import pandas as pd
from attrs import define, field
from Bio import Seq, SeqIO
from Bio.SeqRecord import SeqRecord
from gffutils import FeatureDB
from gffutils.feature import Feature
from goatools.obo_parser import GODag
from sortedcontainers import SortedDict, SortedSet
from torch_geometric.data import download_url

from torchcell.sequence import (  # Genome,
    DnaSelectionResult,
    DnaWindowResult,
    Gene,
    GeneSet,
    calculate_window_bounds,
    calculate_window_bounds_symmetric,
    get_chr_from_description,
    mismatch_positions,
    roman_to_int,
)


@define
class SCerevisiaeGenome:
    data_root: str = field(init=True, repr=False, default="data/sgd/genome")
    # Additional field to store genomes
    genomes: SortedDict = field(init=False)

    def __attrs_post_init__(self) -> None:
        self.genomes = SortedDict()

        # Define the exact location of your .gbff file
        gbff_file_path = os.path.join(
            self.data_root, "ncbi_dataset/data/GCF_000146045.2/genomic.gbff"
        )

        # If the .gbff file doesn't exist, download and extract it
        if not os.path.exists(gbff_file_path):
            print("The .gbff file doesn't exist. Downloading now.")
            self.download_and_extract_genome_files()

        # Read the .gbff file
        self.read_single_gbff_file()

    def read_single_gbff_file(self) -> None:
        # Reusing the gbff_file_path
        gbff_file_path = os.path.join(
            self.data_root, "ncbi_dataset/data/GCF_000146045.2/genomic.gbff"
        )

        # Check if the .gbff file exists
        if not os.path.exists(gbff_file_path):
            print(f"The file {gbff_file_path} does not exist.")
            return

        # Parse the .gbff file and store the result
        for record in SeqIO.parse(gbff_file_path, "genbank"):
            self.genomes[record.id] = record
        print()

    def download_and_extract_genome_files(self) -> None:
        # Full path to the download location
        download_path = os.path.abspath("GCF_000146045.2.zip")

        print(f"Downloading to {download_path}")

        # Download the dataset
        try:
            subprocess.run(
                [
                    "datasets",
                    "download",
                    "genome",
                    "accession",
                    "GCF_000146045.2",
                    "--include",
                    "gbff",
                    "--filename",
                    download_path,
                ]
            )
        except Exception as e:
            print(f"Failed to download: {e}")
            return

        # Check if the zip file is present
        if not os.path.exists(download_path):
            print("The zip file was not downloaded correctly.")
            return

        print("Zip file exists, attempting to unzip.")

        # Unzip the downloaded file
        try:
            with zipfile.ZipFile(download_path, "r") as zip_ref:
                zip_ref.extractall(".")
            print("Unzip operation completed.")
        except Exception as e:
            print(f"Failed to unzip: {e}")
            return

        # Print the directory contents to debug
        for root, dirs, files in os.walk("./GCF_000146045.2"):
            print(root)
            print("Directories:", dirs)
            print("Files:", files)
            print("---")

        # Define the destination directory
        destination_directory = os.path.abspath("data/scerevisiae/genome/s288c/")

        # Create destination directory if it does not exist
        os.makedirs(destination_directory, exist_ok=True)

        print(f"Moving to {destination_directory}")

        # Move the 'ncbi_dataset' to the destination directory
        try:
            shutil.move("./ncbi_dataset", destination_directory)
            print("Move operation completed.")
        except Exception as e:
            print(f"An error occurred while moving the file: {e}")
            return

        # Remove the ZIP file and temporary directory
        try:
            os.remove(download_path)
            print("Cleanup operation completed.")
        except Exception as e:
            print(f"An error occurred while cleaning up: {e}")


def main():
    from dotenv import load_dotenv

    load_dotenv()
    DATA_ROOT = os.getenv("DATA_ROOT")

    genome = SCerevisiaeGenome(
        data_root=osp.join(DATA_ROOT, "data/scerevisiae/genome/s288c")
    )
    pass


if __name__ == "__main__":
    main()



# torchcell/models/deep_set.py
# [[torchcell.models.deep_set]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/models/deep_set.py
# Test file: torchcell/models/test_deep_set.py

import torch
import torch.nn as nn
from torch_scatter import scatter_add, scatter_max, scatter_mean


class SimpleLinearModel(nn.Module):
    def __init__(self, input_dim: int, output_dim: int, scatter: str = "add"):
        super().__init__()
        self.scatter = scatter
        # The main linear layer for the aggregated data
        self.linear = nn.Linear(input_dim, output_dim)

    def forward(self, x, batch):
        # Aggregate the data
        if self.scatter == "add":
            x = scatter_add(x, batch, dim=0)
        elif self.scatter == "mean":
            x = scatter_mean(x, batch, dim=0)
        elif self.scatter == "max":
            x = scatter_max(x, batch, dim=0)

        # Process aggregated data through the linear layer
        x_set = self.linear(x)

        return x_set


def main():
    torch.autograd.set_detect_anomaly(True)

    # Model configuration
    input_dim = 10
    output_dim = 8  # For the aggregated data

    model = SimpleLinearModel(input_dim, output_dim)

    # Dummy data
    x = torch.rand(100, input_dim)
    batch = torch.cat([torch.full((20,), i, dtype=torch.long) for i in range(5)])

    # Forward pass
    x_set = model(x, batch)
    print(x_set.shape)

    # Let's assume you want to predict some values for each set.
    # So, we'll create a dummy target tensor for demonstration purposes.
    target = torch.rand(5, output_dim)

    # Simple mean squared error loss
    criterion = nn.MSELoss()
    loss = criterion(x_set, target)
    print("Loss:", loss.item())

    # Backpropagation
    model.zero_grad()
    loss.backward()
    print("Gradients computed successfully!")


if __name__ == "__main__":
    main()



# torchcell/loader/cpu_experiment_loader
# [[torchcell.loader.cpu_experiment_loader]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/loader/cpu_experiment_loader
# Test file: tests/torchcell/loader/test_cpu_experiment_loader.py



from threading import Thread
from queue import Queue
from typing import Iterable


class CpuExperimentLoader:
    def __init__(self, dataset, batch_size: int, num_workers: int):
        self.dataset = dataset
        self.batch_size = batch_size
        self.load_queue = Queue()
        self.data_queue = Queue(maxsize=num_workers)
        self.workers = []

        # Calculate how many batches are needed
        self.total_batches = (len(dataset) + batch_size - 1) // batch_size

        # Flag to track if close has been called
        # make close idempotent
        self.is_closed = False

        for _ in range(num_workers):
            worker = Thread(
                target=self.worker_function,
                args=(self.load_queue, self.data_queue, self.dataset, self.batch_size),
            )
            worker.start()
            self.workers.append(worker)

        # Initially, load the first few batches depending on the number of workers
        for i in range(min(self.total_batches, num_workers)):
            self.load_queue.put(i)  # Signal with unique batch index

    @staticmethod
    def worker_function(load_queue: Queue, data_queue: Queue, dataset, batch_size: int):
        while True:
            batch_index = load_queue.get()  # Get unique batch index
            if batch_index is None:  # If None signal received, terminate the worker
                break

            start_idx = batch_index * batch_size
            end_idx = min(start_idx + batch_size, len(dataset))
            batch = [dataset[i] for i in range(start_idx, end_idx)]
            data_queue.put(batch)

    def __iter__(self) -> Iterable:
        self.batch_index = 0  # Reset batch index for new iteration
        return self

    def __next__(self):
        if self.batch_index >= self.total_batches:
            self.close()  # Ensure cleanup when iteration is complete
            raise StopIteration

        batch = self.data_queue.get()

        # Prepare next batch if there are more batches to process
        if self.batch_index + len(self.workers) < self.total_batches:
            self.load_queue.put(self.batch_index + len(self.workers))

        self.batch_index += 1
        return batch

    def close(self):
        if not self.is_closed:
            # Send termination signal to each worker
            for _ in self.workers:
                self.load_queue.put(None)
            for worker in self.workers:
                worker.join()  # Wait for all workers to finish
            self.is_closed = True  # Set the flag to prevent repeated cleanup



import json

import torch
from tqdm import tqdm
from transformers import AutoModelForMaskedLM, AutoTokenizer

# Read the data
with open("data/bary_data.json", "r") as f:
    data = json.load(f)
sequences = [entry["sequence"] for entry in data.values()]

# Nucleotide Transformer
tokenizer = AutoTokenizer.from_pretrained(
    "InstaDeepAI/nucleotide-transformer-2.5b-multi-species"
)
model = AutoModelForMaskedLM.from_pretrained(
    "InstaDeepAI/nucleotide-transformer-2.5b-multi-species"
)

# Initialize a list to hold mean sequence embeddings
mean_sequence_embeddings = []

# Loop through sequences and process them one by one
for sequence in tqdm(sequences):
    tokens_ids = tokenizer.encode_plus(
        sequence, return_tensors="pt", truncation=True, padding=True
    )["input_ids"]
    attention_mask = tokens_ids != tokenizer.pad_token_id
    torch_outs = model(
        tokens_ids, attention_mask=attention_mask, output_hidden_states=True
    )

    # Compute the mean embedding for this sequence
    embedding = torch_outs["hidden_states"][-1].detach()
    mean_embedding = torch.sum(
        attention_mask.unsqueeze(-1) * embedding, axis=-2
    ) / torch.sum(attention_mask, axis=-1)
    mean_sequence_embeddings.append(mean_embedding.squeeze().tolist())

# Add the embeddings to the data dictionary
for i, key in enumerate(data.keys()):
    data[key]["embedding"] = mean_sequence_embeddings[i]

# Dump the updated data dictionary to a new JSON file
with open("data/bary_embed.json", "w") as f:
    json.dump(data, f)



# torchcell/neo4j_fitness_query
# [[torchcell.neo4j_fitness_query]]
# https://github.com/Mjvolk3/torchcell/tree/main/torchcell/neo4j_fitness_query
# Test file: tests/torchcell/test_neo4j_fitness_query.py

import attrs
import lmdb
from neo4j import GraphDatabase
import os
from tqdm import tqdm
from attrs import define, field
import os.path as osp


@define
class Neo4jQueryDatabase:
    uri: str
    username: str
    password: str
    root_dir: str
    query: str
    raw_dir: str = field(init=False, default=None)
    env: str = field(init=False, default=None)

    def __attrs_post_init__(self):
        self.raw_dir = osp.join(self.root_dir, "raw", "lmdb")

        if not osp.exists(self.raw_dir):
            os.makedirs(self.raw_dir)
        # Ensure the root_dir exists
        os.makedirs(self.raw_dir, exist_ok=True)
        # Initialize LMDB environment
        self.env = lmdb.open(self.raw_dir, map_size=int(1e6))  # Adjust map_size as needed

        # Process the data
        self.process_data()

    def fetch_data(self):
        driver = GraphDatabase.driver(self.uri, auth=(self.username, self.password))
        with driver.session() as session:
            result = session.run(self.query)
            for record in result:
                print(record)
                yield record

        driver.close()

    def write_to_lmdb(self, key: bytes, value: bytes):
        with self.env.begin(write=True) as txn:
            txn.put(key, value)

    def process_data(self):
        for i, record in tqdm(enumerate(self.fetch_data())):
            # Assuming each record can be uniquely identified and serialized
            # Adjust the key and value encoding as per your data structure
            key = f"record_{i}".encode()  # Example key
            value = str(record).encode()  # Example value serialization

            self.write_to_lmdb(key, value)
        print(f"Finished on {i}")
        self.env.close()


# Example usage
if __name__ == "__main__":
    neo4j_db = Neo4jQueryDatabase(
        uri="bolt://localhost:7687",
        username="neo4j",
        password="torchcell",  # Replace with your actual password
        root_dir="data/torchcell/dmf-2022_02_12",
        query="""
            MATCH (n)
            RETURN n LIMIT 10
        """,
    )

## Get experiments and references
# MATCH (e:Experiment)-[:GenotypeMemberOf]->(g:Genotype)-[:PerturbationMemberOf]->(p:Perturbation {perturbation_type: 'deletion'})
# WITH e, COLLECT(p) AS perturbations
# WHERE ANY(p IN perturbations WHERE p.perturbation_type = 'deletion')
# MATCH (e)-[:Reference]->(ref:ExperimentReference)
# RETURN e, ref
